@@ -191,11 +191,36 @@ func emptyStringTest(m dsl.Matcher) {
 }
 
 //doc:summary Detects redundant conversions between string and []byte
-//doc:tags    style
+//doc:tags    performance
 //doc:before  copy(b, []byte(s))
 //doc:after   copy(b, s)
 func stringXbytes(m dsl.Matcher) {
 	m.Match(`copy($_, []byte($s))`).Report("can simplify `[]byte($s)` to `$s`")
+
+	m.Match(`string($b) == ""`).Where(m["b"].Type.Is(`[]byte`)).Suggest(`len($b) == 0`)
+	m.Match(`string($b) != ""`).Where(m["b"].Type.Is(`[]byte`)).Suggest(`len($b) != 0`)
+
+	m.Match(`len(string($b))`).Where(m["b"].Type.Is(`[]byte`)).Suggest(`len($b)`)
+
+	m.Match(`string($x) == string($y)`).
+		Where(m["x"].Type.Is(`[]byte`) && m["y"].Type.Is(`[]byte`)).
+		Suggest(`bytes.Equal($x, $y)`)
+
+	m.Match(`string($x) != string($y)`).
+		Where(m["x"].Type.Is(`[]byte`) && m["y"].Type.Is(`[]byte`)).
+		Suggest(`!bytes.Equal($x, $y)`)
+
+	m.Match(`$re.Match([]byte($s))`).
+		Where(m["re"].Type.Is(`*regexp.Regexp`) && m["s"].Type.Is(`string`)).
+		Suggest(`$re.MatchString($s)`)
+
+	m.Match(`$re.FindIndex([]byte($s))`).
+		Where(m["re"].Type.Is(`*regexp.Regexp`) && m["s"].Type.Is(`string`)).
+		Suggest(`$re.FindStringIndex($s)`)
+
+	m.Match(`$re.FindAllIndex([]byte($s), $n)`).
+		Where(m["re"].Type.Is(`*regexp.Regexp`) && m["s"].Type.Is(`string`)).
+		Suggest(`$re.FindAllStringIndex($s, $n)`)
 }
 
 //doc:summary Detects strings.Index calls that may cause unwanted allocs
@@ -272,6 +297,8 @@ func badCall(m dsl.Matcher) {
 		Report(`suspicious arg 0, probably meant -1`).At(m["zero"])
 
 	m.Match(`append($_)`).Report(`no-op append call, probably missing arguments`)
+
+	m.Match(`filepath.Join($_)`).Report(`suspicious Join on 1 argument`)
 }
 
 //doc:summary Detects assignments that can be simplified by using assignment operators
@@ -421,3 +448,255 @@ func sliceClear(m dsl.Matcher) {
 		Where(m["zero"].Value.Int() == 0).
 		Report(`rewrite as for-range so compiler can recognize this pattern`)
 }
+
+//doc:summary Detects sync.Map load+delete operations that can be replaced with LoadAndDelete
+//doc:tags    diagnostic experimental
+//doc:before  v, ok := m.Load(k); if ok { m.Delete($k); f(v); }
+//doc:after   v, deleted := m.LoadAndDelete(k); if deleted { f(v) }
+func syncMapLoadAndDelete(m dsl.Matcher) {
+	m.Match(`$_, $ok := $m.Load($k); if $ok { $m.Delete($k); $*_ }`).
+		Where(m.GoVersion().GreaterEqThan("1.15") &&
+			m["m"].Type.Is(`*sync.Map`)).
+		Report(`use $m.LoadAndDelete to perform load+delete operations atomically`)
+}
+
+//doc:summary Detects "%s" formatting directives that can be replaced with %q
+//doc:tags    diagnostic experimental
+//doc:before  fmt.Sprintf(`"%s"`, s)
+//doc:after   fmt.Sprintf(`%q`, s)
+func sprintfQuotedString(m dsl.Matcher) {
+	m.Match(`fmt.Sprintf($s, $*_)`).
+		Where(m["s"].Text.Matches("^`.*\"%s\".*`$") ||
+			m["s"].Text.Matches(`^".*\\"%s\\".*"$`)).
+		Report(`use %q instead of "%s" for quoted strings`)
+}
+
+//doc:summary Detects various off-by-one kind of errors
+//doc:tags    diagnostic
+//doc:before  xs[len(xs)]
+//doc:after   xs[len(xs)-1]
+func offBy1(m dsl.Matcher) {
+	m.Match(`$x[len($x)]`).
+		Where(m["x"].Pure && m["x"].Type.Is(`[]$_`)).
+		Suggest(`$x[len($x)-1]`).
+		Report(`index expr always panics; maybe you wanted $x[len($x)-1]?`)
+
+	// TODO: use $slicing[$i:$*_] form when we'll update go-ruleguard
+	// version so it includes https://github.com/quasilyte/go-ruleguard/pull/284
+
+	m.Match(
+		`$i := strings.Index($s, $_); $_ := $slicing[$i:]`,
+		`$i := strings.Index($s, $_); $_ = $slicing[$i:]`,
+		`$i := bytes.Index($s, $_); $_ := $slicing[$i:]`,
+		`$i := bytes.Index($s, $_); $_ = $slicing[$i:]`).
+		Where(m["s"].Text == m["slicing"].Text).
+		Report(`Index() can return -1; maybe you wanted to do $s[$i+1:]`).
+		At(m["slicing"])
+
+	m.Match(
+		`$i := strings.Index($s, $_); $_ := $slicing[:$i]`,
+		`$i := strings.Index($s, $_); $_ = $slicing[:$i]`,
+		`$i := bytes.Index($s, $_); $_ := $slicing[:$i]`,
+		`$i := bytes.Index($s, $_); $_ = $slicing[:$i]`).
+		Where(m["s"].Text == m["slicing"].Text).
+		Report(`Index() can return -1; maybe you wanted to do $s[:$i+1]`).
+		At(m["slicing"])
+
+	m.Match(
+		`$s[strings.Index($s, $_):]`,
+		`$s[:strings.Index($s, $_)]`,
+		`$s[bytes.Index($s, $_):]`,
+		`$s[:bytes.Index($s, $_)]`).
+		Report(`Index() can return -1; maybe you wanted to do Index()+1`)
+}
+
+//doc:summary Detects slice expressions that can be simplified to sliced expression itself
+//doc:tags    style
+//doc:before  copy(b[:], values...)
+//doc:after   copy(b, values...)
+func unslice(m dsl.Matcher) {
+	m.Match(`$s[:]`).
+		Where(m["s"].Type.Is(`string`) || m["s"].Type.Is(`[]$_`)).
+		Suggest(`$s`).
+		Report(`could simplify $$ to $s`)
+}
+
+//doc:summary Detects Yoda style expressions and suggests to replace them
+//doc:tags    style experimental
+//doc:before  return nil != ptr
+//doc:after   return ptr != nil
+func yodaStyleExpr(m dsl.Matcher) {
+	m.Match(`$constval != $x`).Where(m["constval"].Node.Is(`BasicLit`) && !m["x"].Node.Is(`BasicLit`)).
+		Report(`consider to change order in expression to $x != $constval`)
+	m.Match(`$constval == $x`).Where(m["constval"].Node.Is(`BasicLit`) && !m["x"].Node.Is(`BasicLit`)).
+		Report(`consider to change order in expression to $x == $constval`)
+
+	m.Match(`nil != $x`).Where(!m["x"].Node.Is(`BasicLit`)).
+		Report(`consider to change order in expression to $x != nil`)
+	m.Match(`nil == $x`).Where(!m["x"].Node.Is(`BasicLit`)).
+		Report(`consider to change order in expression to $x == nil`)
+}
+
+//doc:summary Detects unoptimal strings/bytes case-insensitive comparison
+//doc:tags    performance experimental
+//doc:before  strings.ToLower(x) == strings.ToLower(y)
+//doc:after   strings.EqualFold(x, y)
+func equalFold(m dsl.Matcher) {
+	// We specify so many patterns to avoid too generic
+	// patterns that would match things like
+	// `strings.ToLower(x) == strings.ToUpper(y)`
+	// While it could be an EqualFold candidate,
+	// it just looks wrong and should probably be
+	// marked by some other checker.
+
+	// string == patterns
+	m.Match(
+		`strings.ToLower($x) == $y`,
+		`strings.ToLower($x) == strings.ToLower($y)`,
+		`$x == strings.ToLower($y)`,
+		`strings.ToUpper($x) == $y`,
+		`strings.ToUpper($x) == strings.ToUpper($y)`,
+		`$x == strings.ToUpper($y)`).
+		Where(m["x"].Pure && m["y"].Pure && m["x"].Text != m["y"].Text).
+		Suggest(`strings.EqualFold($x, $y)]`).
+		Report(`consider replacing with strings.EqualFold($x, $y)`)
+
+	// string != patterns
+	m.Match(
+		`strings.ToLower($x) != $y`,
+		`strings.ToLower($x) != strings.ToLower($y)`,
+		`$x != strings.ToLower($y)`,
+		`strings.ToUpper($x) != $y`,
+		`strings.ToUpper($x) != strings.ToUpper($y)`,
+		`$x != strings.ToUpper($y)`).
+		Where(m["x"].Pure && m["y"].Pure && m["x"].Text != m["y"].Text).
+		Suggest(`!strings.EqualFold($x, $y)]`).
+		Report(`consider replacing with !strings.EqualFold($x, $y)`)
+
+	// bytes.Equal patterns
+	m.Match(
+		`bytes.Equal(bytes.ToLower($x), $y)`,
+		`bytes.Equal(bytes.ToLower($x), bytes.ToLower($y))`,
+		`bytes.Equal($x, bytes.ToLower($y))`,
+		`bytes.Equal(bytes.ToUpper($x), $y)`,
+		`bytes.Equal(bytes.ToUpper($x), bytes.ToUpper($y))`,
+		`bytes.Equal($x, bytes.ToUpper($y))`).
+		Where(m["x"].Pure && m["y"].Pure && m["x"].Text != m["y"].Text).
+		Suggest(`bytes.EqualFold($x, $y)]`).
+		Report(`consider replacing with bytes.EqualFold($x, $y)`)
+}
+
+//doc:summary Detects suspicious arguments order
+//doc:tags    diagnostic
+//doc:before  strings.HasPrefix("#", userpass)
+//doc:after   strings.HasPrefix(userpass, "#")
+func argOrder(m dsl.Matcher) {
+	m.Match(
+		`strings.HasPrefix($lit, $s)`,
+		`bytes.HasPrefix($lit, $s)`,
+		`strings.HasSuffix($lit, $s)`,
+		`bytes.HasSuffix($lit, $s)`,
+		`strings.Contains($lit, $s)`,
+		`bytes.Contains($lit, $s)`,
+		`strings.TrimPrefix($lit, $s)`,
+		`bytes.TrimPrefix($lit, $s)`,
+		`strings.TrimSuffix($lit, $s)`,
+		`bytes.TrimSuffix($lit, $s)`,
+		`strings.Split($lit, $s)`,
+		`bytes.Split($lit, $s)`).
+		Where((m["lit"].Const || m["lit"].ConstSlice) &&
+			!(m["s"].Const || m["s"].ConstSlice) &&
+			!m["lit"].Node.Is(`Ident`)).
+		Report(`$lit and $s arguments order looks reversed`)
+}
+
+//doc:summary Detects string concat operations that can be simplified
+//doc:tags    style experimental
+//doc:before  strings.Join([]string{x, y}, "_")
+//doc:after   x + "_" + y
+func stringConcatSimplify(m dsl.Matcher) {
+	m.Match(`strings.Join([]string{$x, $y}, "")`).Suggest(`$x + $y`)
+	m.Match(`strings.Join([]string{$x, $y, $z}, "")`).Suggest(`$x + $y + $z`)
+	m.Match(`strings.Join([]string{$x, $y}, $glue)`).Suggest(`$x + $glue + $y`)
+}
+
+//doc:summary Detects manual conversion to milli- or microseconds
+//doc:tags    style experimental
+//doc:before  t.Unix() / 1000
+//doc:after   t.UnixMilli()
+func timeExprSimplify(m dsl.Matcher) {
+	m.Match(`$t.Unix() / 1000`).
+		Where(m.GoVersion().GreaterEqThan("1.17") &&
+			(m["t"].Type.Is(`time.Time`) || m["t"].Type.Is(`*time.Time`))).
+		Suggest("$t.UnixMilli()").
+		Report(`use $t.UnixMilli() instead of $$`)
+
+	m.Match(`$t.UnixNano() * 1000`).
+		Where(m.GoVersion().GreaterEqThan("1.17") &&
+			(m["t"].Type.Is(`time.Time`) || m["t"].Type.Is(`*time.Time`))).
+		Suggest("$t.UnixMicro()").
+		Report(`use $t.UnixMicro() instead of $$`)
+}
+
+//doc:summary Detects exposed methods from sync.Mutex and sync.RWMutex
+//doc:tags    style experimental
+//doc:before  type Foo struct{ ...; sync.Mutex; ... }
+//doc:after   type Foo struct{ ...; mu sync.Mutex; ... }
+func exposedSyncMutex(m dsl.Matcher) {
+	m.Match(`type $x struct { $*_; sync.Mutex; $*_ }`).
+		Where(m["x"].Text.Matches(`^\p{Lu}`)).
+		Report("don't embed sync.Mutex")
+
+	m.Match(`type $x struct { $*_; *sync.Mutex; $*_ }`).
+		Where(m["x"].Text.Matches(`^\p{Lu}`)).
+		Report("don't embed *sync.Mutex")
+
+	m.Match(`type $x struct { $*_; sync.RWMutex; $*_ }`).
+		Where(m["x"].Text.Matches(`^\p{Lu}`)).
+		Report("don't embed sync.RWMutex")
+
+	m.Match(`type $x struct { $*_; *sync.RWMutex; $*_ }`).
+		Where(m["x"].Text.Matches(`^\p{Lu}`)).
+		Report("don't embed *sync.RWMutex")
+}
+
+//doc:summary Detects bad usage of sort package
+//doc:tags    diagnostic experimental
+//doc:before  xs = sort.StringSlice(xs)
+//doc:after   sort.Strings(xs)
+func badSorting(m dsl.Matcher) {
+	m.Match(`$x = sort.IntSlice($x)`).
+		Where(m["x"].Type.Is(`[]int`)).
+		Suggest(`sort.Ints($x)`).
+		Report(`suspicious sort.IntSlice usage, maybe sort.Ints was intended?`)
+
+	m.Match(`$x = sort.Float64Slice($x)`).
+		Where(m["x"].Type.Is(`[]float64`)).
+		Suggest(`sort.Float64s($x)`).
+		Report(`suspicious sort.Float64s usage, maybe sort.Float64s was intended?`)
+
+	m.Match(`$x = sort.StringSlice($x)`).
+		Where(m["x"].Type.Is(`[]string`)).
+		Suggest(`sort.Strings($x)`).
+		Report(`suspicious sort.StringSlice usage, maybe sort.Strings was intended?`)
+}
+
+//doc:summary Detects suspicious reassigment of error from another package
+//doc:tags    diagnostic experimental
+//doc:before  io.EOF = nil
+//doc:after   /* don't do it */
+func externalErrorReassign(m dsl.Matcher) {
+	m.Match(`$pkg.$err = $x`).
+		Where(m["err"].Type.Is(`error`) && m["pkg"].Object.Is(`PkgName`)).
+		Report(`suspicious reassigment of error from another package`)
+}
+
+//doc:summary Detects suspicious empty declarations blocks
+//doc:tags    diagnostic experimental
+//doc:before  var()
+//doc:after   /* nothing */
+func emptyDecl(m dsl.Matcher) {
+	m.Match(`var()`).Report(`empty var() block`)
+	m.Match(`const()`).Report(`empty const() block`)
+	m.Match(`type()`).Report(`empty type() block`)
+}
